/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Annozilla.
 *
 * The Initial Developer of the Original Code is
 * Doug Daniels <rainking@alumni.rice.edu>
 * Portions created by the Initial Developer are Copyright (C) 2003
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   C. Greg Hagerty <cgreg@cgreg.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

/**
 * Constants.
 */
const NS_LOCAL_ANNOTATIONS_SERVER_CID = Components.ID("{32685ae0-1c9d-4862-b022-4f59dd3fa8b6}");
const NS_LOCAL_ANNOTATIONS_SERVER_PROG_ID = "@mozilla.org/local_annotations_server;1";


/**
 * Global vars
 */

// global RDF Service
var gRDFService = Components
    .classes["@mozilla.org/rdf/rdf-service;1"]
    .getService(Components.interfaces.nsIRDFService);

/**
 * A local annotations server, storing its data
 * in RDF and implementing
 * Components.interfaces.nsIAnnotationsServer
 */
function LocalServer() {

    // whether the datasource is loaded
    this.loaded = false;

    var that = this; /* Copy for use inside closure */
    
    // observer to observer loading of datasource
    var localLoadObserver = {
        onBeginLoad: function(aSink) {},
        onInterrupt: function(aSink) {},
        onResume   : function(aSink) {},
        onEndLoad  : function(aSink)
        { 
            that.loaded = true;
        },
        onError    : function(aSink, aStatus, aErrorMsg) { 
            dump("Error loading LocalServer dataSource, " + aErrorMsg + "\n"); 
        }
    };
    
    try {
        // get the preferences directory
        var fileLocator = Components.classes["@mozilla.org/file/directory_service;1"].getService(Components.interfaces.nsIProperties);  
        var file = fileLocator.get("ProfD", Components.interfaces.nsIFile);
        
        // append servers filename
        file.append(LocalServer.LOCAL_ANNOTATIONS_FILENAME);        
        
        // if the file does not exist, create it
        if (!file.exists()) {
            file = LocalServer.createNewAnnotationsFile(file);
        }                
        
        //dump("Localserver datasource: "+LocalServer.getURLSpecFromFile(file)+"\n");
        // get the data source
        this.dataSource = gRDFService.GetDataSource(LocalServer.getURLSpecFromFile(file));
        
        // test whether it's loaded
        var remote = this.dataSource.QueryInterface(Components.interfaces.nsIRDFRemoteDataSource);
        if (remote.loaded) {
            this.loaded = true;
        } 
        else {
            // Attach an observer to the loading of the RDF
            var sink =
                this.dataSource.QueryInterface(Components.interfaces.nsIRDFXMLSink);
            sink.addXMLSinkObserver(localLoadObserver);
        }        
    }
    catch(e) {
        dump("LocalServer Loading Error: " + e + "\n");
    }
}

LocalServer.LOCAL_ANNOTATIONS_FILENAME = "localAnnotations.rdf";
LocalServer.EMPTY_ANNOTATIONS_RDF = 
    "<?xml version=\"1.0\"?>\n" +
    "<r:RDF \n"+
    "xmlns:d=\"http://purl.org/dc/elements/1.0/\" \n" +
    "xmlns:t=\"http://www.w3.org/2001/03/thread#\" \n" +
    "xmlns:atr=\"http://www.w3.org/2001/12/attributions/ns#\" \n" +
    "xmlns:http=\"http://www.w3.org/1999/xx/http#\" \n" +
    "xmlns:p=\"xmlns:http://www.w3.org/2000/08/palm56/addr#\" \n" +
    "xmlns:a=\"http://www.w3.org/2000/10/annotation-ns#\" \n" +
    "xmlns:r=\"http://www.w3.org/1999/02/22-rdf-syntax-ns#\">\n</r:RDF>\n";

// File read/write operations
LocalServer.NS_RDONLY      = 0x01;
LocalServer.NS_WRONLY      = 0x02;
LocalServer.NS_CREATE_FILE = 0x08;
LocalServer.NS_TRUNCATE    = 0x20;

// user -- read, write; group -- read; other -- read
LocalServer.DEFAULT_FILE_PERMS = 00400 | 00200 | 00040 | 00004;

LocalServer.URI = "urn:localserver";
LocalServer.USERNAME = "noUser";
LocalServer.PASSWORD = "noPass";

LocalServer.DEFAULT_ENCODING = "UTF-8";

/**
 * Arbitrary namespace for the body of an Annotation.
 */
LocalServer.LOCALSERVER_NS = "http://cnx.rice.edu/annotations/localServer#";
LocalServer.ANNOT_NS = "http://www.w3.org/2000/10/annotation-ns#";
LocalServer.ANNOT_TYPE_NS = "http://www.w3.org/2000/10/annotationType#";
LocalServer.RDF_NS   = "http://www.w3.org/1999/02/22-rdf-syntax-ns#";
LocalServer.DC_NS    = "http://purl.org/dc/elements/1.0/";
LocalServer.THREAD_NS= "http://www.w3.org/2001/03/thread#";
LocalServer.HTTP_NS = "http://www.w3.org/1999/xx/http#";

/**
 * Literals to parse from RDF.
 */
LocalServer.REQUIRED_METADATA_LITERALS = [   
    { namespace: LocalServer.DC_NS, property: "creator"},
    { namespace: LocalServer.ANNOT_NS, property: "created"},
    { namespace: LocalServer.DC_NS, property: "date"},
    { namespace: LocalServer.ANNOT_NS, property: "context"} ];

LocalServer.REQUIRED_METADATA_RESOURCES = [
    { namespace: LocalServer.RDF_NS, property: "type"},
    { namespace: LocalServer.ANNOT_NS, property: "annotates"} ];


/**
 * RDF Parsing for Annotation namespace.
 */
LocalServer.ANNOTATION_TYPE = LocalServer.ANNOT_NS + "Annotation";

/**
 * RDF Parsing for Reply namespace.
 */
LocalServer.REPLY_TYPE = LocalServer.THREAD_NS + "Reply";

LocalServer.prototype = {
    
    get url() { return LocalServer.URI; },

    get username() { return LocalServer.USERNAME; },
    set username(name) { throw Components.Exception("You cannot set the LocalServer username."); },

    get password() { return LocalServer.PASSWORD; },
    set password(name) { throw Components.Exception("You cannot set the LocalServer password.");  }
};

LocalServer.prototype.QueryInterface = function(iid) {
    if (!iid.equals(Components.interfaces.nsIAnnotationsServer) &&
        !iid.equals(Components.interfaces.nsISupports)) {
            throw Components.results.NS_ERROR_NO_INTERFACE;
        }
    return this;
}

/**
 * No extra initialization necessary for local server.
 *  ++so overload with hack to load a different annotations file!
 *  @param url=="urn:localserver" or local annotations filename
 *  @param username (anything)
 *  @param password (anything)
 */
var gPathInvariantAnnotations = false;
LocalServer.prototype.init = function(url, username, password) {

    if (url != "urn:localserver") {
    try {
         dump("LocalServer: Initializing datasource "+url+"\n");
        
         var prefManager = Components.classes["@mozilla.org/preferences-service;1"]
                  .getService(Components.interfaces.nsIPrefBranch);
         if (prefManager.getBoolPref(
                  "LocalAnnotationsServer.path_invariant_annotations")) {
             gPathInvariantAnnotations = true;             
             dump(" PathInvariantAnnotations: on\n");
         } else {
             gPathInvariantAnnotations = false;
             dump(" PathInvariantAnnotations: off\n");             
         }

         file = Components.classes["@mozilla.org/file/local;1"]
                     .createInstance(Components.interfaces.nsILocalFile);
         
         file.initWithPath(url);
        
        // if the file does not exist, create it
        if (!file.exists()) {
            file = LocalServer.createNewAnnotationsFile(file);
        }                
        
        // get the data source
        this.dataSource = gRDFService.GetDataSource(LocalServer.getURLSpecFromFile(file));
        
        // test whether it's loaded
        var remote = this.dataSource.QueryInterface(Components.interfaces.nsIRDFRemoteDataSource);
        if (remote.loaded) {
            this.loaded = true;
        } 
        else {
            this.loaded = false;

            var that = this; /* Copy for use inside closure */
            
            // observer to observer loading of datasource
            var localLoadObserver = {
                onBeginLoad: function(aSink) {},
                onInterrupt: function(aSink) {},
                onResume   : function(aSink) {},
                onEndLoad  : function(aSink)
                { 
                    that.loaded = true;
                },
                onError    : function(aSink, aStatus, aErrorMsg) { 
                    dump("Error loading LocalServer dataSource, " + aErrorMsg + "\n"); 
                }
            };
            
            // Attach an observer to the loading of the RDF
            var sink =
                this.dataSource.QueryInterface(Components.interfaces.nsIRDFXMLSink);
            sink.addXMLSinkObserver(localLoadObserver);
        }        
    }
    catch(e) {
        dump("LocalServer Loading Error: " + e + "\n");
    }}
}



/*
 * Allow refresh of datasource on demand
 *  in case of external modification
 *
 * @param waitp whether to wait or not
 */
//debug to console 
//var gConsoleService = Components.classes['@mozilla.org/consoleservice;1']
//    .getService(Components.interfaces.nsIConsoleService);

LocalServer.prototype.refreshDataSource = function (waitp) {
//  gConsoleService.logStringMessage ("Refreshing localserver datasource");
  dump("Localserver: Refreshing datasource\n");
  this.dataSource.Refresh(waitp); 
}

/**
 * Post a new annotation to the server.  
 *
 * @param data nsIMetadata for a new annotation 
 * @param bodyText string text of annotation body
 * @param observer 
 */
LocalServer.prototype.postAnnotation =
function(data, bodyText, encoding, observer, target, postWindow) {
    var annotationUriDigitsLength = 10;

    if (! this.loaded) {
        observer.postAnnotationOnFailure
            ("LocalServer RDF Data Source still loading.  Please try again shortly.",
             target,
             postWindow);
        return;
    }

    /* ensure that the metadata argument has the required fields,
       and load them in */
    var unionObj = null;
    try {
        unionObj = LocalServer.loadPostData(data);
    }
    catch(e) {
        observer.postAnnotationOnFailure
            ("Insufficient data to post annotation: " + e.toString(),
             target,
             postWindow);
        return;
    }

    // extract literals and resources from union object
    var literals = unionObj.lit;
    var resources = unionObj.res;
    var primary = unionObj.primary;
    var secondary = unionObj.secondary;
    var annotation = Components.classes["@mozilla.org/annotation;1"].createInstance(Components.interfaces.nsIAnnotation);
    var bareNumbers = LocalServer.generateRandomNumberString(annotationUriDigitsLength);
    var annotURI = "urn:annot" + bareNumbers;
    var bodyURI = "urn:body" + bareNumbers;

    // init the annotation
    annotation.initialize(LocalServer.URI, annotURI, bodyURI, null);

    var annotResource = gRDFService.GetResource(annotURI);

    // get the url that we're annotating
    var annotates = resources[LocalServer.ANNOT_NS + "annotates"].value;
    
    // assert our annotation under this url
    this.dataSource.Assert(gRDFService.GetResource(annotates),
                           gRDFService.GetResource(LocalServer.LOCALSERVER_NS + "annotatedBy"),
                           annotResource, true);
        
    // assert the resources
    for(prop in resources) {
        var resource = resources[prop];
      
        this.dataSource.Assert(annotResource,
                               gRDFService.GetResource(resource.namespace + resource.property),
                               gRDFService.GetResource(resource.value), true);
        annotation.data.setResourceProperty(resource.namespace, resource.property, 
                                            resource.value);
    }
    
    // assert the literals
    for (prop in literals) {
        var literal = literals[prop];
        this.dataSource.Assert(annotResource,
                               gRDFService.GetResource(literal.namespace + literal.property),
                               gRDFService.GetLiteral(literal.value), true);
        annotation.data.setLiteralProperty(literal.namespace, literal.property, 
                                           literal.value);
    }

    this.dataSource.Assert(annotResource,
                           gRDFService.GetResource(LocalServer.RDF_NS + "type"),
                           gRDFService.GetResource(secondary), true);
    annotation.data.setResourceProperty(LocalServer.RDF_NS, "type", 
                                        secondary);
    /* RDF Datasource can't serialize two rdf:type fields!
    this.dataSource.Assert(annotResource,
                           gRDFService.GetResource(LocalServer.RDF_NS + "type"),
                           gRDFService.GetResource(primary), true);
*/
    annotation.data.setResourceProperty(LocalServer.RDF_NS, "type", 
                                        primary);  

    // assert the body
    var bodyResource = gRDFService.GetResource(bodyURI);
    
    // if there is no bodyText, use empty string
    if (! bodyText) {
        bodyText = "";
    }
    this.dataSource.Assert(annotResource,
                           gRDFService.GetResource(LocalServer.ANNOT_NS + "body"),
                           bodyResource, true);
    this.dataSource.Assert(bodyResource,
                           gRDFService.GetResource(LocalServer.LOCALSERVER_NS + "bodyText"),
                           gRDFService.GetLiteral(bodyText), true);                   

    // set the bodyText, for efficiency
    annotation.setBodyText(bodyText);

    // flush the data source
    var remote = this.dataSource.QueryInterface(Components.interfaces.nsIRDFRemoteDataSource);
    remote.Flush();

    observer.postAnnotationOnSuccess(annotation, target, postWindow);
}


/**
 * Load the annotations on this server for the given URI.
 *
 * @param uri string URI to load annotations for
 * @param observer 
 */
LocalServer.prototype.loadAnnotations = function(uri, observer, target) {
    
    if (! this.loaded) {
        observer.loadAnnotationsOnFailure(LocalServer.URI, uri, 
                                          "LocalServer RDF Data Source still loading.  Please try again shortly.",
                                          target);
    }
    else {
        var annotations = this._loadAnnotationsForURI(uri);
        observer.loadAnnotationsOnSuccess(LocalServer.URI, uri, 
                                          annotations.length, annotations, target);
    }
}

/**
 * Delete the annotation at the specified uri.
 *
 * @param uri string uri to delete
 * @param observer 
 */
LocalServer.prototype.deleteAnnotation = function(annotURI, observer, target) {
    var annotates = null;

    if (! this.loaded) {
        observer.deleteAnnotationOnFailure(annotURI, "LocalServer RDF Data Source still loading.  Please try again shortly.", target);
        return;
    }

    // get URIs and Resources for Annotation and its Body
    // relies on uris being urn:annot<num> and urn:body<num>
    var bodyURI = annotURI.replace("annot", "body");
    var annotResource = gRDFService.GetResource(annotURI);
    
    // Determine if this annotation actually exists, by checking whether
    // whether there are any resources originating in the annotation URI
    var hasAnnotation = false;
    var arcsEnum = this.dataSource.ArcLabelsOut(annotResource);
    while (arcsEnum.hasMoreElements()) {
        hasAnnotation = true;
        break;
    }
    if (!hasAnnotation)
    {
        observer.deleteAnnotationOnFailure(annotURI, "Annotation does not exist", target);
        return;
    }

    var bodyResource = gRDFService.GetResource(bodyURI);

    // unassert the body text
    var bodyTextPropertyResource =
        gRDFService.GetResource(LocalServer.LOCALSERVER_NS + "bodyText");
    var bodyTextValueNode = this.dataSource.GetTarget(bodyResource, bodyTextPropertyResource, true);
    this.dataSource.Unassert(bodyResource, bodyTextPropertyResource, bodyTextValueNode);

    // unassert all arcs from annotation
    var arcs = this.dataSource.ArcLabelsOut(annotResource);
    while (arcs.hasMoreElements()) {
        var arc = arcs.getNext().QueryInterface(Components.interfaces.nsIRDFResource);
        var arcValueNode = this.dataSource.GetTarget(annotResource, arc, true);
        // if this is annotates, store it
        if (arc.Value == LocalServer.ANNOT_NS + "annotates") {
            annotates = arcValueNode.QueryInterface(Components.interfaces.nsIRDFResource).Value;
        }
        this.dataSource.Unassert(annotResource, arc, arcValueNode);
    }

    // unassert the annotatedBy property
    if (! annotates) {
        throw Components.Exception("Malformed RDF: couldn't get annotates property");
    }
    var uriResource = gRDFService.GetResource(annotates);
    var annotatedByProperty = gRDFService.GetResource(LocalServer.LOCALSERVER_NS + "annotatedBy");
    this.dataSource.Unassert(uriResource, annotatedByProperty, annotResource);

    // flush the data source
    var remote = this.dataSource.QueryInterface(Components.interfaces.nsIRDFRemoteDataSource);
    remote.Flush();

    observer.deleteAnnotationOnSuccess(annotURI, target);
}

/**
 * Update the annotation passed in on the server.  
 *
 * @param annotation nsIAnnotation
 * @param observer nsIAnnotationsServerObserver
 */
LocalServer.prototype.updateAnnotation = function(annotation, observer, target, updateWindow) {
    var oldAnnotates = null;

    // ensure that annotationURI is properly defined
    if ( (!annotation.annotationURI) || (annotation.annotationURI == "") ) {
        observer.updateAnnotationOnFailure
            ("",
             "annotationURI for parameter annotation is not defined.",
             target,
             updateWindow);
        return;
    }

    // ensure that bodyURI is properly defined
    if ( (!annotation.bodyURI) || (annotation.bodyURI == "") ) {
        observer.updateAnnotationOnFailure
            (annotation.annotationURI,
             "bodyURI for parameter annotation is not defined.",
             target,
             updateWindow);
        return;
    }

    // get URIs and Resources for Annotation and its Body
    var annotResource = gRDFService.GetResource(annotation.annotationURI);
    var bodyResource = gRDFService.GetResource(annotation.bodyURI);
    
    /* ensure that the metadata argument has the required fields,
       and load them in */
    var unionObj = null;
    try {
        unionObj = LocalServer.loadPostData(annotation.data);
    }
    catch(e) {
        observer.updateAnnotationOnFailure
           (annotation.annotationURI,
            "Insufficient data to update annotation: " + e.toString(),
            target,
            updateWindow);
        return;
    }

    // extract literals and resources from union object
    var literals = unionObj.lit;
    var resources = unionObj.res;
    var primary = unionObj.primary;
    var secondary = unionObj.secondary;

    // unassert all the old arcs out of the annotation except body
    var arcs = this.dataSource.ArcLabelsOut(annotResource);
    while(arcs.hasMoreElements()) {
        var arc = arcs.getNext().QueryInterface(Components.interfaces.nsIRDFResource);
        var arcValueNode = this.dataSource.GetTarget(annotResource, arc, true);
        // if this is annotates, store it
        if (arc.Value == LocalServer.ANNOT_NS + "annotates") {
          oldAnnotates = arcValueNode.QueryInterface(Components.interfaces.nsIRDFResource).Value;
        }
        else if (arc.Value == (LocalServer.ANNOT_NS + "body")) {
          continue;
        }
        this.dataSource.Unassert(annotResource, arc, arcValueNode);
    }

    // assert the new resources
    for(prop in resources) {
        var resource = resources[prop];
        
        this.dataSource.Assert(annotResource,
                               gRDFService.GetResource(resource.namespace + resource.property),
                               gRDFService.GetResource(resource.value), true);
    }
    
    // assert the new literals
    for (prop in literals) {
        var literal = literals[prop];
        this.dataSource.Assert(annotResource,
                               gRDFService.GetResource(literal.namespace + literal.property),
                               gRDFService.GetLiteral(literal.value), true);
    }
/*
    this.dataSource.Assert(annotResource,
                           gRDFService.GetResource("http://www.w3.org/1999/02/22-rdf-syntax-ns#type"),
                           gRDFService.GetResource(primary), true);
*/
    this.dataSource.Assert(annotResource,
                           gRDFService.GetResource("http://www.w3.org/1999/02/22-rdf-syntax-ns#type"),
                           gRDFService.GetResource(secondary), true);

    // check whether we need to change the annotatedBy property
    if (! oldAnnotates) {
      throw Components.Exception("Malformed RDF: couldn't get annotates property");
    }
    var newAnnotates = annotation.data.getProperty(LocalServer.ANNOT_NS, "annotates");
    if (oldAnnotates != newAnnotates) {
        // unassert the old annotates
        var oldResource = gRDFService.GetResource(oldAnnotates);
        var annotatedByProperty = gRDFService.GetResource(LocalServer.LOCALSERVER_NS + "annotatedBy");
        this.dataSource.Unassert(oldResource, annotatedByProperty, annotResource);

        // assert the new annotates
        var newResource = gRDFService.GetResource(newAnnotates);
        this.dataSource.Assert(newResource, annotatedByProperty, annotResource, true);
    }

    // do the body if necessary
    var bodyText = annotation.getBodyText();
    if (bodyText) {
        var bodyTextPropertyResource = gRDFService.GetResource(LocalServer.LOCALSERVER_NS + 
                                                               "bodyText");
        var bodyTextValueNode = this.dataSource.GetTarget(bodyResource, 
                                                          bodyTextPropertyResource, true);
        this.dataSource.Change(bodyResource, bodyTextPropertyResource, bodyTextValueNode,
                               gRDFService.GetLiteral(bodyText));

    }
    
    // flush the data source
    var remote = this.dataSource.QueryInterface(Components.interfaces.nsIRDFRemoteDataSource);
    remote.Flush();
        
    observer.updateAnnotationOnSuccess(annotation, target, updateWindow);
}

/**
 * Load the replies for an annotation.
 *
 * @param annotation nsIAnnotation
 * @param observer nsIAnnotationsServerObserver
 */
LocalServer.prototype.getReplies = function(annotation, observer, target) {
    observer.getRepliesOnFailure
      (annotation,
       "getReplies is not implemented on the Local Server",
       target);
}

/**
 * Reply to an annotation.
 *
 * @param original nsIAnnotation original annotation
 * @param reply nsIMetadata reply data
 * @param observer nsIAnnotationsServerObserver observer
 */
LocalServer.prototype.replyToAnnotation =
function(original, metadata, bodyText, encoding, observer, target, replyWindow) {
    observer.replyToAnnotationOnFailure
        (original,
         "replyToAnnotation is not implemented on the Local Server",
         target,
         replyWindow);
}

/**
 * Load the body of an annotation into the "body" Metadata
 * field of the annotation.
 *
 * @param annotation nsIAnnotation to load body for
 * @param observer nsIAnnotationsObserver
 */
LocalServer.prototype.loadAnnotationBody = function(annotation, observer) {
    if ( (!annotation.bodyURI) || (annotation.bodyURI == "") ) {
        observer.loadAnnotationBodyOnFailure(annotation, 
                                             "Annotation bodyURI is not defined");
        return;
    }
    var bodyText = this.getAnnotationBodyText(annotation);
    
    if (! bodyText) {
        observer.loadAnnotationBodyOnFailure(annotation,
                                             "Unable to find body text in local annotations RDF file.");
        return;
    }
    
    annotation.setBodyText(bodyText);
    observer.loadAnnotationBodyOnSuccess(annotation, bodyText);
}
    
    

/**
 * toString.
 *
 * @return String rep of this
 */
LocalServer.prototype.toString = function() { return "[LocalServer]"; }


// PRIVATE methods

LocalServer.createNewAnnotationsFile = function(file) {
     
    // initialize local file
    var localFile = Components.classes["@mozilla.org/file/local;1"].createInstance(Components.interfaces.nsILocalFile);
    
    localFile.initWithPath(file.path);
    
    if (localFile.exists() && 
        (localFile.isDirectory() || !localFile.isWritable())) {
        throw Components.results.NS_ERROR_FAILURE;                
    }
        
    // write flags
    var flags = LocalServer.NS_WRONLY | LocalServer.NS_CREATE_FILE | LocalServer.NS_TRUNCATE;

    // get a file output stream
    var fileStream = Components.classes["@mozilla.org/network/file-output-stream;1"].createInstance(Components.interfaces.nsIFileOutputStream);
    fileStream.init(localFile, flags, LocalServer.DEFAULT_FILE_PERMS, 0);

    fileStream.write(LocalServer.EMPTY_ANNOTATIONS_RDF, LocalServer.EMPTY_ANNOTATIONS_RDF.length);
    fileStream.close();

    return localFile;
}

/**
 * Gets a URL spec from a file, working
 * for whichever version of mozilla you
 * happen to be running.
 *
 * @param file nsIFile to get url spec for
 * @return nsIFileURL
 */
LocalServer.getURLSpecFromFile = function(file) {
    
    if (Components.classes["@mozilla.org/network/io-service;1"] && Components.interfaces.nsIIOService) {
        
        var ioService = Components.classes["@mozilla.org/network/io-service;1"].getService(Components.interfaces.nsIIOService);
        var fileHandler = ioService.getProtocolHandler("file").QueryInterface(Components.interfaces.nsIFileProtocolHandler);

        // new version
        if (fileHandler && fileHandler.getURLSpecFromFile) {
            return fileHandler.getURLSpecFromFile(file);
        }
        // old version in ioService
        else {
            return ioService.getURLSpecFromFile(file);
        }
    }
    // really old version
    else {
        var file_url = Components.classes["@mozilla.org/network/standard-url;1"].createInstance(Components.interfaces.nsIFileURL);
        file_url.file = file;
        return file_url.spec;
    }
}


/**
 * Generates a random integral number,
 * possibly with leading zeros.
 *
 * @param length length of number
 */
LocalServer.generateRandomNumberString = function(length) {
    var uri = "";

    for (var i=0; i < length; i++) {
        // get an integer of uri_length digits
        rand = Math.floor(Math.random() * 10);
        uri += rand;
    }
    
    return uri;
}

/**
 * Loads the values from an nsIAnnotationMetadata
 * object, and tests that the necessary values are defined.
 *
 * @param meta nsIAnnotationMetadata to load data from
 * @throws Exception if a required field is undefined
 * @return requiredData with values filled in if all fields exist
 */
LocalServer.loadPostData = function(meta) {
    var literals = {};
    var resources = {};
    
    if (meta.getPrimaryType() == undefined) {
        throw Components.Exception("Metadata is missing primary type");
    }
    if (meta.getSecondaryType() == undefined) {
        throw Components.Exception("Metadata is missing secondary type");
    }
    
    // check for required literals
    for (var i = 0; i < LocalServer.REQUIRED_METADATA_LITERALS; i++) {
        var literal = LocalServer.REQUIRED_METADATA_LITERALS[i];
        var aValue = meta.getProperty(literal.namespace, literal.property);

        if (aValue) {
            literals[literal.namespace + literal.property] = { namespace: literal.namespace, property: literal.property, value: aValue };
        }
        else {
            
        }
    }
    
    // check for required resources
    for (var i = 0; i < LocalServer.REQUIRED_METADATA_RESOURCES; i++) {
        var resource = LocalServer.REQUIRED_METADATA_RESOURCES[i];
        var aValue = meta.getProperty(resource.namespace, resource.property);

        if (aValue) {
            resources[resource.namespace + resource.property] = 
            { namespace: resource.namespace, property: resource.property, value: aValue };
        }
        else {
            throw Components.Exception("Metadata is missing required resource " + 
                                       resource.namespace + resource.property + "\n");
        }
    }

    // enumerate the metadata to add additional items
    var properties = {};
    var count = {};
    meta.getAllProperties(count, properties);
    for (var i = 0; i < properties.value.length; i++) {
        var datum = properties.value[i];
        if (datum.type == datum.LITERAL) {
            if (! literals[datum.namespace + datum.property]) {
                
                literals[datum.namespace + datum.property] = 
                { namespace: datum.namespace, property: datum.property, value: datum.value };
            }
        }
        else {
            if (! resources[datum.namespace + datum.property]) {
                resources[datum.namespace + datum.property] = 
                { namespace: datum.namespace, property: datum.property, value: datum.value };
            }
        }
    }

    return { lit: literals, res: resources,
             primary: meta.getPrimaryType(), secondary: meta.getSecondaryType() };
}


/**
 * Given an rdfDataSource, a resourceEnum (subset of the dataSource) and the resources and literals to extract,
 * creates <code>Annotation<code>s with these resources and literals in the metadata
 * of the Annotation.  
 *
 * @param rdfDataSrc nsIRDFDataSource to mine for annotations
 * @param resourceEnum nsISimpleEnumerator subset of resources to build annotations for
 * @param resourceNames names of fields of this object will be parsed as resources from the RDF
 * @param literalNames names of fields of this object will be parsed as literals from the RDF
 * @param rdfType one of either ANNOTATION_TYPE or REPLY_TYPE
 * @param optional list of annotations to add to
 * @return Annotation[] of Annotations
 */
LocalServer.prototype.buildAnnotationsFromResourceEnum =
function(rdfDataSrc, resourceEnum, rdfType, serverName, annotations) 
{
    if (!annotations) { annotations = []; }

    while(resourceEnum.hasMoreElements()) {

        // get resource
        var res = resourceEnum.getNext();
        res = res.QueryInterface(Components.interfaces.nsIRDFResource);

        var annotURI = res.Value;
        var bodyURI  = LocalServer.getRDFResource(res, LocalServer.ANNOT_NS, "body", rdfDataSrc);
        
        var annotation = Components.classes["@mozilla.org/annotation;1"].createInstance(Components.interfaces.nsIAnnotation);
        annotation.initialize(serverName, annotURI, bodyURI, null);

        // view all the arcs out
        var arcs = rdfDataSrc.ArcLabelsOut(res);
        while (arcs.hasMoreElements()) {
            var arc = arcs.getNext().QueryInterface(Components.interfaces.nsIRDFResource);
            if (arc.Value == LocalServer.ANNOT_NS + "body") {
                continue;
            }

            var splitValue = LocalServer.splitResourceString(arc.Value);
            
            if (splitValue) {
                // try to get it as a resource
                var value = LocalServer.getRDFResource(res, splitValue.namespace, splitValue.property, rdfDataSrc);
                // if not a resource, try a literal
                if (! value) {
                    value = LocalServer.getRDFLiteral(res, splitValue.namespace, splitValue.property, rdfDataSrc);
                    if (! value) {
                        value = "";
                    }
                    annotation.data.setLiteralProperty(splitValue.namespace, splitValue.property, value);
                }
                else {
                    annotation.data.setResourceProperty(splitValue.namespace, splitValue.property, value);
                }
            }
        }
        
        // TODO Currently the Local Server supports only annotations
        // (not replies). The "primary type" isn't stored in the RDF, either,
        // so add it manually
        annotation.data.setResourceProperty (LocalServer.RDF_NS, "type", rdfType);

        // additionally, load and set the annotation bodyText
        // as an optimization
        var bodyText = this.getAnnotationBodyText(annotation);
        if (bodyText) {
          annotation.setBodyText(bodyText);
        }

        annotations.push(annotation);
    }

    return annotations;
}


/**
 * Splits a namespace + property Resource string into
 * namespace and property.
 *
 * @param resourceString string of namespace + property
 * @return object with namespace and property fields
 */
LocalServer.splitResourceString = function(resourceString) {
    var i;
    for (i = (resourceString.length - 1); i >= 0; i--) {
        var charAtI = resourceString.charAt(i);
        if (charAtI == '#' || charAtI == '/' || charAtI == ':') {
            return { namespace: resourceString.substring(0, i+1),
                                    property: resourceString.substring(i+1) };
        }
    }
    
    // couldn't find it -- return null
    return null;
}



/**
 * Get a property from an RDF Literal.
 *
 * @param resource RDF resource
 * @param ns String namespace
 * @param property String property to retrieve
 * @param rdfDataSrc RDF Data Source
 * @return value of literal or undefined if not found
 *  
 */
LocalServer.getRDFLiteral = function(resource, ns, property, rdfDataSrc) {
    return LocalServer.getRDFValue (resource, ns, property, rdfDataSrc,
                             Components.interfaces.nsIRDFLiteral);
}

/**
 * Get a property from an RDF resource.
 *
 * @param resource RDF resource
 * @param ns String namespace
 * @param property String property to retrieve
 * @param rdfDataSrc RDF Data Source
 * @return value of property, or undefined if not found
 *  
 */
LocalServer.getRDFResource = function(resource, ns, property, rdfDataSrc) {
    return LocalServer.getRDFValue (resource, ns, property, rdfDataSrc,
                             Components.interfaces.nsIRDFResource);
}

/**
 * Get an RDF Value.
 *
 * @param resource RDF resource
 * @param ns String namespace
 * @param property String property to retrieve
 * @param rdfDataSrc RDF Data Source
 * @param iface interface to query
 * @return value of resource or undefined if property not found
 *  
 */
LocalServer.getRDFValue = function(resource, ns, property, rdfDataSrc, iface) {
    try {
        var RDFNode =
        rdfDataSrc.GetTarget(resource, gRDFService.GetResource(ns + property), true);
        RDFNode = RDFNode.QueryInterface(iface);
        RDFNode = RDFNode.Value;
        return RDFNode;
    } catch (e) {
        return null;
    }
}


/**
 * Make a deep clone of an array of annotations.
 *
 * @param oldAnnots array of annotations
 * @return deep clone
 */
LocalServer._cloneAnnotations = function(oldAnnots) {
      var newAnnots = [];
      for (var i = 0; i < oldAnnots.length; i++) {
          newAnnots.push(oldAnnots[i].clone());
      }
      return newAnnots;
}

/**
 * Loads the local annotations for a uri.
 *
 * @param uri string uri for which to load annotations
 * @return array of nsIAnnotation for uri
 */
LocalServer.prototype._loadAnnotationsForURI = function(uri) {
    var annotations = [];
    var resourceEnum = this.dataSource.GetTargets
        (
         gRDFService.GetResource(uri),
         gRDFService.GetResource(LocalServer.LOCALSERVER_NS + "annotatedBy"),
         true
    );
    
    annotations = this.buildAnnotationsFromResourceEnum
        (this.dataSource,
         resourceEnum,
         LocalServer.ANNOTATION_TYPE,
         LocalServer.URI, annotations);

    if (gPathInvariantAnnotations) {
        // THE FOLLOWING CODE ALLOWS FOR PATH INVARIANT ANNOTATIONS - OPTIONAL?
        var shortURI = uri.lastIndexOf("/")+1;
        if ((shortURI > 0) && (shortURI < uri.length)) {
            // && path invariance annotation sets option   
            shortURI = uri.substr(shortURI); 
            dump("Loading local annotations for [short uri] "+shortURI+"\n");
            resourceEnum = this.dataSource.GetTargets
                (gRDFService.GetResource(shortURI),
                 gRDFService.GetResource(LocalServer.LOCALSERVER_NS+"annotatedBy"),
                 true);
            
            annotations = this.buildAnnotationsFromResourceEnum
                (this.dataSource, resourceEnum,
                 LocalServer.ANNOTATION_TYPE,
                 LocalServer.URI, annotations);
        }
    }
  return annotations;
}

LocalServer.prototype.getAnnotationBodyText = function(annotation) {
  return LocalServer.getRDFLiteral(gRDFService.GetResource(annotation.bodyURI),
                                   LocalServer.LOCALSERVER_NS,
                                   "bodyText",
                                   this.dataSource);
}


/**************************
 * XPCOM stuff
 **************************/

/**
 * Factory for nsLocalAnnotationsServer.
 */
var serverFactory = new Object();

serverFactory.createInstance = function (outer, iid) {
    if (outer != null) {
        throw Components.results.NS_ERROR_NO_AGGREGATION;
    }
    
    if (!iid.equals(Components.interfaces.nsISupports)) {
        throw Components.results.NS_ERROR_INVALID_ARG;
    }

    return new LocalServer();
}


/**
 * Module for nsLocalAnnotationsServer.
 */
var nsLocalAnnotationsServerModule = new Object();

nsLocalAnnotationsServerModule.registerSelf =
    function (compMgr, fileSpec, location, type) {
        compMgr = compMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
        dump("***** Local Annotations Server Service -- registering self ******\n");
        debug("***** Local Annotations Server Service -- registering self ******\n");
        compMgr.registerFactoryLocation(NS_LOCAL_ANNOTATIONS_SERVER_CID,
                                        "Local Annotations Server Service",
                                        NS_LOCAL_ANNOTATIONS_SERVER_PROG_ID, 
                                        fileSpec,
                                        location,
                                        type);
    }

nsLocalAnnotationsServerModule.unregisterSelf = function(compMgr, fileSpec, location) {
    compMgr = compMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
    compMgr.unregisterFactoryLocation(NS_LOCAL_ANNOTATIONS_SERVER_CID, fileSpec);
}

nsLocalAnnotationsServerModule.getClassObject = function (compMgr, cid, iid) {
    if (cid.equals(NS_LOCAL_ANNOTATIONS_SERVER_CID)) {
        return serverFactory;
    }
    else {
        throw Components.results.NS_ERROR_NO_INTERFACE;  
    }
}

nsLocalAnnotationsServerModule.canUnload = function(compMgr) {
    return true;
}

/* entrypoint */
function NSGetModule(compMgr, fileSpec) {
    return nsLocalAnnotationsServerModule;
}
